const Firmata = require('firmata-io').Firmata;
const SerialPort = require('serialport');
const log = require('minilog')('firmata-link');
require('minilog').enable();
const JsonRpcWs = require('json-rpc-ws');


class FirmataLink {
    constructor () {
        this.firmata = null;
        this.serialportOptions = {baudRate: 57600, bufferSize: 256};

        this.serverPort = 2020;
        this.listTimeoutTime = 5000;

        this.rpcServer = new JsonRpcWs.createServer();
    }

    exposeMethods () {
        this.rpcServer.expose('scan', (params, reply) => {
            this.listDevices()
                .then(deviceList => {
                    if (Object.keys(deviceList) == 0) {
                        reply("Device not found", null);
                    } else {
                        reply(null, deviceList);
                    }
                });
            
        });
        this.rpcServer.expose('connect', (params, reply) => {
            if (!params.peripheralId) {
                reply('peripheralId is null', null);
            }
            this.connectPort(params.peripheralId)
                .then(firmata => {
                    reply(null, this._getBoardProperty(firmata));
                })
                .catch(reason => {
                    log.error(reason);
                    reply(reason, null);
                });
        });
        this.rpcServer.expose('disconnect', (params, reply) => {
            try {
                this.closeBoardOn();
                reply(null, params);
            } catch (e) {
                reply(e, params);
            }
        });
        this.rpcServer.expose('getBoardState', (params, reply) => {
            reply(null, this.getBoardStateOn(params.peripheralId));
        });
        this.rpcServer.expose('digitalWrite', (params, reply) => {
            try {
                this.digitalWrite(params.peripheralId, params.pin, params.value);
                reply(null, params);
            } catch (e) {
                reply(e, params);
            }
        });
        this.rpcServer.expose('pwmWrite', (params, reply) => {
            try {
                this.pwmWrite(params.peripheralId, params.pin, params.value);
                reply(null, params);
            } catch (e) {
                reply(e, params);
            }
        });
        this.rpcServer.expose('pinMode', (params, reply) => {
            try {
                this.pinMode(params.peripheralId, params.pin, params.mode);
                reply(null, params);
            } catch (e) {
                reply(e, params);
            }
        });
        this.rpcServer.expose('sendString', (params, reply) => {
            try {
                this.sendString(params.peripheralId, params.str);
                reply(null, params);
            } catch (e) {
                reply(e, params);
            }
        });
    }

    startServer () {
        this.exposeMethods();
        return new Promise((resolve, reject) => {

            this.rpcServer.start({port: this.serverPort}, err => {
                if (err) {
                    reject(err);
                }
                log.info(`FirmataLink started on ws://localhost:${this.serverPort}`);
                resolve(this.rpcServer);
            });
        });
    }

    stopServer () {
        this.rpcServer.stop();
    }

    closeBoardOn () {
        if (!this.firmata) return;
        if (this.firmata.transport.isOpen) {
            this.firmata.transport.close();
        }
        log.info(`Close board on ${this.firmata.transport.path}`);
        
        delete this.firmata;
        this.firmata = null;
    }

    release () {
        this.stopServer();
        this.closeBoardOn();
    }

    _openTransport (peripheralId) {
        return new Promise((resolve, reject) => {
            const newPort = new SerialPort(peripheralId, this.serialportOptions,
                err => {
                    if (err) {
                        log.error(err);
                        reject(err);
                    }
                });
            resolve(newPort);
        });
    }

    _attachFirmata (port) {
        return new Promise((resolve, reject) => {
            const firmata =
                new Firmata(
                    port,
                    {
                        skipCapabilities: false,
                        reportVersionTimeout: 1500 /* MCU AutoSend=1000[ms], Query=1500[ms], must be < listTimeoutTime */
                    },
                    err => {
                        if (err) {
                            // Called on 'error'
                            log.info(err);
                            reject(err);
                        } else {
                            // Called on 'ready'
                            resolve(firmata);
                        }
                    });
        });
    }

    _openBoardOn (peripheralId) {
        return new Promise(async (resolve, reject) => {
            const port = await this._openTransport(peripheralId);
            try {
                const firmata = await this._attachFirmata(port);
                resolve(firmata);
            } catch (err) {
                port.close();
                reject(err);
            }
        });
    }


    async listDevices() {
        const ports = await SerialPort.list();
        let deviceList = {};
        
        ports.forEach(port => {
            deviceList[port.path] = {
                name: port.path,
                peripheralId: port.path
            };
        });
        
        log.debug(JSON.stringify(ports, null, "    "));
        return (deviceList);
    }


    _initializeBoard (firmata) {
        for (let i = 0; i < firmata.pins.length; i++) {
            const ain = firmata.analogPins.indexOf(i);
            if (ain >= 0) {
                firmata.reportAnalogPin(ain, 1);
            } else {
                firmata.reportDigitalPin(i, 1);
            }
        }
    }

    async connectPort (peripheralId) {
        if (this.firmata) {
            if (this.firmata.transport.path == peripheralId) {
                return (this.firmata);
            }
            
            /* close port */
            if (this.firmata.transport.isOpen) {
                this.firmata.transport.close();
            }
            delete this.firmata;
            this.firmata = null;
        }
        
        this.firmata = await this._openBoardOn(peripheralId);
        this._initializeBoard(this.firmata);
        
        log.info(`Connect to ${this.firmata.transport.path} : ${this.firmata.firmware.name}` +
            ` v.${this.firmata.firmware.version.major}.${this.firmata.firmware.version.minor}`);
        return this.firmata;
    }

    getBoardStateOn (peripheralId) {
        const boardState = {};
        if (!this.firmata) return boardState;
        boardState.MODES = this.firmata.MODES;
        boardState.pins = this.firmata.pins;
        boardState.analogPins = this.firmata.analogPins;
        boardState.transport = {
            path: this.firmata.transport.path,
            baudRate: this.firmata.transport.baudRate,
            isOpen: this.firmata.transport.isOpen
        };
        return boardState;
    }

    _getBoardProperty (firmata) {
        return Object.assign({}, firmata,
            {
                name: firmata.transport.path,
                peripheralId: firmata.transport.path,
                transport: {
                    path: firmata.transport.path,
                    baudRate: firmata.transport.baudRate,
                    isOpen: firmata.transport.isOpen
                }
            }
        );
    }

    digitalWrite (peripheralId, pin, value) {
        if (!this.firmata) throw new Error(`Board not found on ${peripheralId}`);
        this.firmata.digitalWrite(pin, value);
    }

    pwmWrite (peripheralId, pin, value) {
        if (!this.firmata) throw new Error(`Board not found on ${peripheralId}`);
        this.firmata.pwmWrite(pin, value);
    }

    pinMode (peripheralId, pin, mode) {
        if (!this.firmata) throw new Error(`Board not found on ${peripheralId}`);
        this.firmata.pinMode(pin, mode);
    }

    sendString (peripheralId, str) {
        if (!this.firmata) throw new Error(`Board not found on ${peripheralId}`);
        this.firmata.sendString(str);
    }
}

module.exports = FirmataLink;
